from marshmallow import Schema, fields


class ExcleSchema(Schema):
    excel_id = fields.Integer()
    name = fields.String()
    import_time = fields.DateTime(format='%Y-%m-%d %H:%M:%S')
    status = fields.Integer()
    excel_type = fields.String()
    extend = fields.String()







import re, os, time, random, xlrd, xlutils.copy, json, openpyxl, shutil
from .utils import is_int, json_serial
from datetime import datetime, date


class ErrorChecker(object):
    config = [
    ]

    def __init__(self, id, error_path, param):
        from .models import get_session, ExcelModel
        db_session = get_session()
        self.db_session = db_session
        excel = db_session.query(ExcelModel).filter(ExcelModel.excel_id == id).first()
        self.excel = excel
        self.error_path = error_path
        self.param = param

    @staticmethod
    def key_handle(key):
        return re.sub(r'\s', '', key).lower()

    def check(self, row_dict, index):
        configs = self.config
        error_list = []
        for config in configs:
            method = config[1]
            error = self.__getattribute__(method)(name=config[0], error_str=config[2],
                                                  row_dict=row_dict, index=index)
            if error:
                if isinstance(error, list):
                    error_list.extend(error)
                else:
                    error_list.append(error)
        return error_list

    def check_date(self, name, error_str, row_dict, index):
        key = self.key_handle(name)
        _date = row_dict.get(key)
        # print(_date)
        if isinstance(_date, str):
            try:
                a = datetime.strptime(_date, '%Y-%m-%d %H:%M:%S')
                return None
            except Exception as e:
                return '第{index}行{name}{error_str}'.format(index=index, name=name, error_str=error_str)
        elif isinstance(_date, float) and len(str(_date)) == 7:
            return None
        elif isinstance(_date, datetime):
            return None
        else:
            return '第{index}行{name}{error_str}'.format(index=index, name=name, error_str=error_str)

    def check_datetime(self, name, error_str, row_dict, index):
        key = self.key_handle(name)
        _datetime = row_dict.get(key)
        if isinstance(_datetime, str):
            try:
                datetime.strptime(_datetime, '%Y-%m-%d %H:%M:%S')
                return None
            except:
                return '第{index}行{name}{error_str}'.format(index=index, name=name, error_str=error_str)
        elif isinstance(_datetime, float) and len(str(_datetime)) == 17:
            return None
        elif isinstance(_datetime, datetime):
            return None
        else:
            return '第{index}行{name}{error_str}'.format(index=index, name=name, error_str=error_str)

    def check_float(self, name, error_str, row_dict, index):
        key = self.key_handle(name)
        _float = row_dict.get(key)
        if isinstance(_float, float):
            return None
        else:
            return '第{index}行{name}{error_str}'.format(index=index, name=name, error_str=error_str)

    def check_str(self, name, error_str, row_dict, index):
        key = self.key_handle(name)
        _str = row_dict.get(key)
        if isinstance(_str, str):
            if _str.strip():
                return None
            else:
                return '第{index}行{name}{error_str}'.format(index=index, name=name, error_str=error_str)
        elif _str is None:
            return '第{index}行{name}{error_str}'.format(index=index, name=name, error_str=error_str)
        else:
            return None

    def check_int(self, name, error_str, row_dict, index):
        key = self.key_handle(name)
        num = row_dict.get(key)
        if is_int(num):
            return None
        return '第{index}行{name}{error_str}'.format(index=index, name=name, error_str=error_str)

    def as_json(self):
        """考虑xls和xlsx的不同"""
        path = self.excel.path
        # print(path)
        if path.split('.')[2] == 'xls':
            excel_open = xlrd.open_workbook(path)
            table = excel_open.sheet_by_index(0)
            rows = table.nrows
            old_excel_title = table.row_values(0)
            excel_title = []
            for title in old_excel_title:
                real_title = self.key_handle(title)
                excel_title.append(real_title)
            json_list = []
            for i in range(2, rows):
                row_dict = dict(zip(excel_title, table.row_values(i)))
                row_dict['error'] = self.check(row_dict, i + 1)
                row_dict['is_use'] = '0'
                json_list.append(row_dict)
        else:
            excel_open = openpyxl.load_workbook(path)
            sheets = excel_open.sheetnames
            sheet0 = sheets[0]
            table = excel_open[sheet0]
            rows = table.rows
            # print(next(rows))
            old_excel_title = [col.value for col in list(rows)[0]]
            excel_title = []
            for title in old_excel_title:
                real_title = self.key_handle(title)
                excel_title.append(real_title)
            json_list = []
            i = 0
            rows = table.rows
            for row in rows:
                i = i + 1
                if i <= 2:
                    continue
                row_value = [col.value for col in row]
                row_dict = dict(zip(excel_title, row_value))  # []
                row_dict['error'] = self.check(row_dict, i)
                row_dict['is_use'] = '0'
                json_list.append(row_dict)

        json_str = json.dumps(json_list, default=json_serial, ensure_ascii=False)
        self.excel.content = json_str
        self.excel.status = 1
        self.db_session.flush()

    def add_error(self):
        path = self.excel.path
        time_str = str(time.time())
        random_str = str(random.randint(1, 10000))
        if path.split('.')[2] == 'xls':
            file_save_name = time_str + random_str + '.xls'
            real_path = os.path.join(self.error_path, file_save_name)
            shutil.copyfile(path, real_path)
            rb = xlrd.open_workbook(real_path, formatting_info=True)
            wb = xlutils.copy.copy(rb)
            sheet = rb.sheet_by_index(0)
            cols = sheet.ncols
            write_sheet = wb.get_sheet(0)
            write_sheet.write(0, cols, 'ERROR')
            contents = json.loads(self.excel.content)
            for i, content in enumerate(contents):
                write_sheet.write(i + 2, cols, str(content.get('error'))[1:-1])
        else:
            file_save_name = time_str + random_str + '.xlsx'
            real_path = os.path.join(self.error_path, file_save_name)
            shutil.copyfile(path, real_path)
            wb = openpyxl.load_workbook(real_path)
            sheets = wb.sheetnames
            sheet0 = sheets[0]
            table = wb[sheet0]
            cols = len(list(table.columns))
            table.cell(row=1, column=cols + 1).value = 'Error'
            contents = json.loads(self.excel.content)
            for i, content in enumerate(contents):
                table.cell(row=i + 3, column=cols + 1).value = str(content.get('error'))[1:-1]
        wb.save(real_path)
        self.excel.error_path = real_path
        self.excel.status = 2
        self.db_session.flush()

    def start(self):
        try:
            self.as_json()
            self.add_error()
        except Exception as e:
            print(e)
            self.excel.status = 3
            self.db_session.flush()



from datetime import datetime
from app import db


def get_session():
    return db.create_scoped_session(options={
        'autoflush': False,
        'autocommit': True
    })


class ExcelModel(db.Model):
    """
    excel_id : 主键
    name: 名称
    path: 路径
    content: 内容,text,包含error
    error_path: 包含错误的excel下载地址
    import_time: 导入时间
    excel的处理状态
    status 状态
    extend 拼接参数的传递
    """
    __tablename__ = 'hs_excel'

    excel_id = db.Column(db.Integer, autoincrement=True, primary_key=True)
    name = db.Column(db.String(128), nullable=True)
    path = db.Column(db.String(256), nullable=True)
    excel_type = db.Column(db.String(50), nullable=True)
    content = db.Column(db.Text, nullable=True)
    error_path = db.Column(db.String(256), nullable=True)
    import_time = db.Column(db.DateTime, default=datetime.now())
    status = db.Column(db.Integer, default=0)
    extend = db.Column(db.String(256), nullable=True)


class MaterialTemporary(db.Model):
    # 物料临时表
    __tablename__ = 'material_temporary'

    material_id = db.Column(db.Integer, primary_key=True)
    excel_id = db.Column(db.Integer, nullable=True)
    material_name = db.Column(db.NVARCHAR(500), nullable=True, default="")  # 物种名称
    material_code = db.Column(db.NVARCHAR(500), nullable=True, default="")  # 物种编码
    stock_unit = db.Column(db.NVARCHAR(50), nullable=True, default="")  # 库存单位
    pr_unit = db.Column(db.NVARCHAR(50), nullable=True, default="")  # 采购单位
    unit_stock2pr_rate = db.Column(db.DECIMAL(9, 4), nullable=True, default=1.00)  # 采购库存转换率
    spec = db.Column(db.NVARCHAR(50), nullable=True, default="")  # 规格型号
    error = db.Column(db.Integer, default=0)



import os, time, random, logging, copy, json, datetime
from flask import current_app, jsonify, Blueprint, request, send_file, make_response
from .utils import allowed_file
from .schemy import ExcleSchema
from urllib.parse import quote
from sqlalchemy import func

hs_excel = Blueprint('hs_excel', __name__)

logger = logging.getLogger(__name__)


@hs_excel.route('/import/', methods=['POST', 'GET', 'DELETE'])
def get_excel():
    """获得excel,并对excel进行基础的格式校验"""
    from .models import get_session, ExcelModel
    from .task import check_error,check
    excel_type = request.args.get('excel_type')
    db_session = get_session()
    if not excel_type:
        return jsonify({'message': '请拼接excel_type'}), 400
    if request.method == 'POST':
        base_path = current_app.config.get('EXCEL_PATH')
        import_path = os.path.join(base_path, 'import')
        error_path = os.path.join(base_path, 'error')
        file = request.files['file']
        if not (file and allowed_file(filename=file.filename)):
            return jsonify({'message': 'excel格式不正确,文件上传失败'}), 400

        # 这里有个方法要对excel进行基础校验，主要是看看能不能解析出来
        time_str = str(time.time())
        random_str = str(random.randint(1, 10000))
        if file.filename.split('.')[1] == 'xls':
            file_save_name = time_str + random_str + '.xls'
        else:
            file_save_name = time_str + random_str + '.xlsx'
        file_path = os.path.join(import_path, file_save_name)
        file.save(file_path)
        param = dict(request.args)
        excel = ExcelModel()
        excel.name = file.filename
        excel.path = file_path
        excel.excel_type = excel_type
        extend = copy.deepcopy(param)
        extend.pop('excel_type')
        excel.extend = json.dumps(extend)
        db_session.add(excel)
        db_session.flush()
        ## 这个地方放异步任务
        check_error.delay(excel.excel_id, param, error_path)
        # check(excel.excel_id,param,error_path)
        schema = ExcleSchema(many=False)
        result = schema.dump(excel).data
        return jsonify(result)
    # 获取上传的所有excel的列表，导入时间倒序排序，带分页
    elif request.method == 'GET':
        page = request.args.get('page', 1)
        per_page = request.args.get('per_page', 10)
        start_time = request.args.get('start_time')
        end_time = request.args.get('end_time')
        query = db_session.query(ExcelModel).filter(ExcelModel.excel_type == excel_type).order_by(
            ExcelModel.import_time)
        if start_time and end_time:
            start_time = datetime.datetime.strptime(start_time, '%Y-%m-%d') - datetime.timedelta(days=1)
            end_time = datetime.datetime.strptime(end_time, '%Y-%m-%d') + datetime.timedelta(days=1)
            query = query.filter(ExcelModel.import_time.between(start_time, end_time))
        paginate = query.paginate(page=int(page), per_page=per_page)
        items = paginate.items
        total = paginate.total
        schema = ExcleSchema(many=True)
        data = schema.dump(items).data
        result = {'data': data,
                  'paging': {'page': int(page),
                             'per_page': int(per_page),
                             'total_number': int(total)}}
        # print(1111111111)
        return jsonify(result)

    else:
        excel_ids = request.get_json()
        excels = db_session.query(ExcelModel).filter(ExcelModel.excel_id.in_(excel_ids)).all()
        for excel in excels:
            db_session.delete(excel)
            path = excel.path
            error_path = excel.error_path
            try:
                os.remove(path)
                os.remove(error_path)
            except Exception as e:
                logger.error(e)
        db_session.flush()
        return jsonify(None), 204


@hs_excel.route('/status/', methods=['GET'])
def get_status():
    excel_id = request.args.get('excel_id')
    if not excel_id:
        return jsonify({'message': 'excel_id必传'}), 400
    from .models import get_session, ExcelModel
    db_session = get_session()
    excel = db_session.query(ExcelModel).filter(ExcelModel.excel_id == excel_id).first()
    if excel:
        return jsonify({'status': excel.status})
    else:
        return {'message': '该文件已被删除'}, 400


@hs_excel.route('/one/excel/', methods=['GET'])
def get_one_excel():
    """获取单个上传的excel"""
    excel_id = request.args.get('excel_id')
    # print(11111111111111)
    if not excel_id:
        return jsonify({'message': 'excel_id必传'}), 400
    from .models import get_session, ExcelModel
    db_session = get_session()
    excel = db_session.query(ExcelModel).filter(ExcelModel.excel_id == excel_id).first()
    path = excel.path
    file_name = excel.name
    # file = send_file(path)
    # response = make_response(file)
    # response.headers["Content-Disposition"] = "attachment; filename={}".format(file_name.encode().decode('latin-1'))
    # filename = FILE_PATH + filename
    response = make_response(send_file(path))
    basename = os.path.basename(file_name)
    response.headers["Content-Disposition"] = \
        "attachment;" \
        "filename*=UTF-8''{utf_filename}".format(
            utf_filename=quote(basename.encode('utf-8'))
        )
    return response


@hs_excel.route('/error/export/', methods=['GET'])
def get_error_excel():
    """下载带错误的excel文件"""
    excel_id = request.args.get('excel_id')
    if not excel_id:
        return jsonify({'message': 'excel_id必传'}), 400
    from .models import get_session, ExcelModel
    db_session = get_session()
    excel = db_session.query(ExcelModel).filter(ExcelModel.excel_id == excel_id).first()
    path = excel.error_path
    file_name = quote(excel.name)
    # file = send_file(path)
    # response = make_response(file)
    # response.headers['Content-Disposition'] += \
    #     "; filename*=utf-8''%s" % (file_name)
    # response.headers["Content-Disposition"] = "attachment; filename={}".format(file_name.encode().decode('latin-1'))
    # return response
    response = make_response(send_file(path))
    basename = os.path.basename(file_name)
    response.headers["Content-Disposition"] = \
        "attachment;" \
        "filename*=UTF-8''{utf_filename}".format(
            utf_filename=quote(basename.encode('utf-8'))
        )
    return response


